"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

ModalBox subclass:#InfoBox
       instanceVariableNames:'formLabel textLabel buttonPanel okButton okAction
			      acceptReturnAsOK'
       classVariableNames:'InfoBitmap'
       poolDictionaries:''
       category:'Views-DialogBoxes'
!

InfoBox comment:'

COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

$Header: /cvs/stx/stx/libwidg/InfoBox.st,v 1.10 1994-12-28 13:53:57 claus Exp $
written Spring/Summer 89 by claus
'!

!InfoBox class methodsFor:'documentation'!

documentation
"
    this class implements a pop-up box to show an information message.
    It has a single ok-Button, which closes the box.
    Also entering return has (by default) the same effect as pressing
    the ok-button.
    InfoBox is a superclass of some other boxes - see WarningBox, YesNoBox etc.
    most of them simply redefine the icon shown in the upper left or
    add buttons.

    instance variables:

	formLabel        <Label>        shows a bitmap (warning, question-mark)

	textLabel        <Label>        shows the boxes text

	buttonPanel      <PanelView>    contains the button(s)

	okButton         <Button>       the ok-Button

	okAction         <Block>        the action to be performed when ok is pressed,
					or return is pressed.

	acceptReturnAsOK <Boolean>      if true, pressing the return-key counts
					as if ok was pressed. Default is true.



    InfoBoxes are created with:

	aBox := InfoBox title:'some title'.

    and shown with:

	aBox showAtPointer
    or
	aBox open

    The default box shows 'yes' in its button; this can be changed with:

	aBox okText:'some string'.

    the boxes bitmap-image can be changed with:

	aBox form:aForm

    (the name 'form:' is historical - any bitmap or image is allowed).
"
!

examples
"
    Examples:

	|aBox|
	aBox := InfoBox title:'a simple Info'.
	aBox showAtPointer.

	|aBox|
	aBox := InfoBox title:'a simple Info'.
	aBox open.

	|aBox|
	aBox := InfoBox title:'a simple Info'.
	aBox showAt:0@0.

	|aBox|
	aBox := InfoBox title:'a simple Info'.
	aBox acceptReturnAsOK:false.
	aBox showAtPointer.

	|aBox|
	aBox := InfoBox title:'Press ''YES'' to continue\(or type return)' withCRs.
	aBox okText:'YES'.
	aBox showAtPointer.

	|aBox|
	aBox := InfoBox title:'another one'.
	aBox form:(Form fromFile:'SBrowser.xbm').
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'a nice one'.
	aBox form:(Image fromFile:'bitmaps/garfield.gif').
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'a nice one'.
	aBox form:((Image fromFile:'bitmaps/garfield.gif') magnifyBy:0.3@0.3).
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'a nice one'.
	aBox form:((Image fromFile:'bitmaps/garfield.gif') magnifyBy:0.3@0.3).
	aBox formLabel level:-1.
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'a nice one'.
	aBox form:((Image fromFile:'bitmaps/garfield.gif') magnifyBy:0.3@0.3).
	aBox formLabel level:1.
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'another nice one'.
	aBox form:((Image fromFile:'bitmaps/garfield.gif') magnifyBy:0.3@0.3).
	aBox formLabel level:1.
	aBox textLabel font:(Font family:'helvetica' face:'bold' style:'roman' size:20).
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'a nice one'.
	aBox form:((Image fromFile:'bitmaps/garfield.gif') magnifyBy:0.3@0.3).
	aBox formLabel borderWidth:2.
	aBox formLabel borderColor:Color red.
	aBox showAtPointer

	|aBox|
	aBox := InfoBox title:'start printing'.
	aBox form:(Image fromFile:'bitmaps/ljet3.xpm').
	aBox formLabel level:2.
	aBox okText:'print'.
	aBox showAtPointer
"
! !

!InfoBox class methodsFor:'styles'!

updateStyleCache
    |img|

    img := StyleSheet at:'informationBoxIcon'.
    img notNil ifTrue:[InfoBitmap := img on:Display].
! !

!InfoBox class methodsFor:'icon bitmap'!

iconBitmap
    "return the bitmap shown as icon in my instances.
     The form is cached and reused, for faster opening."

    InfoBitmap isNil ifTrue:[
	InfoBitmap := (Image fromFile:'bitmaps/Information.xbm') on:Display 
    ].
    ^ InfoBitmap
! !

!InfoBox class methodsFor:'instance creation'!

title:titleString
    "create a new infoBox with title, aTitleString"

    ^ (self new) title:titleString
! !

!InfoBox methodsFor:'initialization'!

initialize
    |mm|

    super initialize.

    label := 'Info'.

    mm := ViewSpacing.

    acceptReturnAsOK := true.

    formLabel := Label in:self.
    self initFormBitmap.
    formLabel borderWidth:0.
    formLabel origin:(mm @ mm).

    textLabel := Label label:'Information' in:self.
    textLabel borderWidth:0.
    textLabel origin:[(mm + formLabel widthIncludingBorder + mm) 
		      @ 
		      mm].

    buttonPanel := HorizontalPanelView in:self.
    buttonPanel origin:(0.0 @ 1.0)
		corner:(1.0 @ 1.0).
    buttonPanel leftInset:mm;
	       rightInset:mm;
		 topInset:(font height * 2 + mm + mm) negated;
	      bottomInset:mm.

    buttonPanel layout:#spread.
    buttonPanel borderWidth:0.

    okButton := Button okButtonIn:buttonPanel.
    okButton action:[
		       okButton turnOffWithoutRedraw.
		       self okPressed 
		    ].
!

initFormBitmap
    "setup the bitmap shown in the upper left -
     extracted into a separate method for easier redefinition
     in subclasses"

    formLabel form:(self class iconBitmap)
! !

!InfoBox methodsFor:'realization'!

positionOffset
    "return the delta, by which the box should be displayed
     from the mouse pointer. Value returned here makes center of
     okButton appear under the cursor"

    buttonPanel setChildPositionsIfChanged.
    ^ (okButton originRelativeTo:self) + (okButton extent // 2)
! !

!InfoBox methodsFor:'accessing'!

acceptReturnAsOK:aBoolean
    "turn on/off interpretation of return-key as ok.
     Default is on"

    acceptReturnAsOK := aBoolean.
    okButton isReturnButton:aBoolean.
!

form:aForm
    "define a form to be displayed left of the title
     - usually an exclamation-mark"

    formLabel form:aForm.
    formLabel resize.
    self resize
!

textLabel
    "return the textLabel = can be used to change its appearance"

    ^ textLabel
!

formLabel
    "return the formLabel = can be used to change its appearance"

    ^ formLabel
!

title:aString
    "set the boxes title string"

    aString ~= textLabel label ifTrue:[
	textLabel label:aString.
	textLabel resize.
	self resize
    ]
!

title
    "return the boxes title string"

    ^ textLabel label
!

okButton
    "return the okButton"

    ^ okButton
!

okAction:aBlock
    "define the action to be performed when ok is pressed"

    okAction := aBlock
!

okText:aString
    "define the text in the ok-button"

    aString ~= okButton label ifTrue:[
	okButton label:aString.
	okButton resize.
	self resize
    ]
! !

!InfoBox methodsFor:'queries'!

preferedExtent 
    "return my prefered extent"

    |w h extra|

    formLabel resize.
    textLabel resize.

    w := margin + 
	 ViewSpacing + 
	 formLabel widthIncludingBorder + 
	 ViewSpacing + textLabel width + ViewSpacing +
	 margin.
    w := w max:(okButton preferedExtent x + (ViewSpacing * 2)).
    h := ViewSpacing
	 + ((formLabel heightIncludingBorder) max:(textLabel height))
	 + ViewSpacing + ViewSpacing + ViewSpacing + ViewSpacing
	 + okButton heightIncludingBorder
	 + ViewSpacing.

    extra := 0 "margin * 2".
    ^ (w + extra) @ (h + extra)
! !

!InfoBox methodsFor:'user interaction'!

hideAndEvaluate:aBlock
    "make myself invisible and evaluate aBlock"

    self hide.
    aBlock notNil ifTrue:[aBlock value]
!

okPressed
    "user pressed ok-button; make myself invisible and if an action was
     specified do it"

    self hideAndEvaluate:okAction
!

keyPress:aKey x:x y:y
    "return-key dublicates ok-function"

    acceptReturnAsOK ifTrue:[
	(aKey == #Return) ifTrue:[self okPressed]
    ]
! !
